@@ -10,8 +10,8 @@ from .logging import logging, setup_logging, QuietError
 from . import fs, CURRENT_FS_REV
 from .backends.pool import BackendPool
 from .block_cache import BlockCache
-from .common import (get_backend_cachedir, get_seq_no, get_backend_factory,
-                     load_params, save_params)
+from .common import (get_seq_no, get_backend_factory, load_params, save_params,
+                     is_mounted)
 from .daemonize import daemonize
 from .database import Connection
 from .inode_cache import InodeCache
@@ -24,7 +24,6 @@ import _thread
 import argparse
 import faulthandler
 import llfuse
-import itertools
 import os
 import platform
 import subprocess
@@ -38,11 +37,6 @@ import time
 import shutil
 import atexit
 
-try:
-    from systemd.daemon import notify as sd_notify
-except ImportError:
-    sd_notify = None
-
 log = logging.getLogger(__name__)
 
 def install_thread_excepthook():
@@ -85,6 +79,12 @@ def main(args=None):
     if not os.path.exists(options.mountpoint):
         raise QuietError('Mountpoint does not exist.', exitcode=36)
 
+    # Check if fs is mounted on this computer
+    # This is not foolproof but should prevent common mistakes
+    if is_mounted(options.storage_url):
+        raise QuietError('File system already mounted elsewhere on this '
+                         'machine.', exitcode=40)
+
     if options.threads is None:
         options.threads = determine_threads(options)
 
@@ -116,13 +116,12 @@ def main(args=None):
         import pstats
         prof = cProfile.Profile()
 
-    backend_factory = get_backend_factory(options.storage_url, options.backend_options,
-                                          options.authfile, options.compress)
+    backend_factory = get_backend_factory(options)
     backend_pool = BackendPool(backend_factory)
     atexit.register(backend_pool.flush)
 
     # Get paths
-    cachepath = get_backend_cachedir(options.storage_url, options.cachedir)
+    cachepath = options.cachepath
 
     # Retrieve metadata
     with backend_pool() as backend:
@@ -158,6 +157,7 @@ def main(args=None):
     operations = fs.Operations(block_cache, db, max_obj_size=param['max_obj_size'],
                                inode_cache=InodeCache(db, param['inode_gen']),
                                upload_event=metadata_upload_thread.event)
+    block_cache.fs = operations
     metadata_upload_thread.fs = operations
 
     with ExitStack() as cm:
@@ -173,22 +173,22 @@ def main(args=None):
             llfuse.close(unmount=unmount_clean)
         cm.callback(unmount)
 
-        if options.fg:
+        if options.fg or options.systemd:
             faulthandler.enable()
             faulthandler.register(signal.SIGUSR1)
         else:
             if stdout_log_handler:
                 logging.getLogger().removeHandler(stdout_log_handler)
-            global crit_log_fh
-            crit_log_fh = open(os.path.join(options.cachedir, 'mount.s3ql_crit.log'), 'a')
-            faulthandler.enable(crit_log_fh)
-            faulthandler.register(signal.SIGUSR1, file=crit_log_fh)
+            crit_log_fd = os.open(os.path.join(options.cachedir, 'mount.s3ql_crit.log'),
+                                  flags=os.O_APPEND|os.O_CREAT|os.O_WRONLY, mode=0o644)
+            faulthandler.enable(crit_log_fd)
+            faulthandler.register(signal.SIGUSR1, file=crit_log_fd)
             daemonize(options.cachedir)
 
         mark_metadata_dirty(backend, cachepath, param)
 
         block_cache.init(options.threads)
-        cm.callback(block_cache.destroy)
+        cm.callback(block_cache.destroy, options.keep_cache)
 
         metadata_upload_thread.start()
         cm.callback(metadata_upload_thread.join)
@@ -198,19 +198,20 @@ def main(args=None):
         cm.callback(commit_thread.join)
         cm.callback(commit_thread.stop)
 
-        if options.upstart:
-            os.kill(os.getpid(), signal.SIGSTOP)
-        if sd_notify is not None:
-            sd_notify('READY=1')
-            sd_notify('MAINPID=%d' % os.getpid())
-
         exc_info = setup_exchook()
         workers = 1 if options.single else None # use default
 
+        if options.systemd:
+            import systemd.daemon
+            systemd.daemon.notify('READY=1')
+
         if options.profile:
-            prof.runcall(llfuse.main, workers)
+            ret = prof.runcall(llfuse.main, workers)
         else:
-            llfuse.main(workers)
+            ret = llfuse.main(workers)
+
+        if ret is not None:
+            raise RuntimeError('Received signal %d, terminating' % (ret,))
 
         # Allow operations to terminate while block_cache is still available
         # (destroy() will be called again when from llfuse.close(), but at that
@@ -357,13 +358,19 @@ def get_metadata(backend, cachepath):
 
     seq_no = get_seq_no(backend)
 
+    # When there was a crash during metadata rotation, we may end up
+    # without an s3ql_metadata object.
+    meta_obj_name = 's3ql_metadata'
+    if meta_obj_name not in backend:
+        meta_obj_name += '_new'
+
     # Check for cached metadata
     db = None
     if os.path.exists(cachepath + '.params'):
         param = load_params(cachepath)
         if param['seq_no'] < seq_no:
             log.info('Ignoring locally cached metadata (outdated).')
-            param = backend.lookup('s3ql_metadata')
+            param = backend.lookup(meta_obj_name)
         elif param['seq_no'] > seq_no:
             raise QuietError("File system not unmounted cleanly, run fsck!",
                              exitcode=30)
@@ -371,7 +378,7 @@ def get_metadata(backend, cachepath):
             log.info('Using cached metadata.')
             db = Connection(cachepath + '.db')
     else:
-        param = backend.lookup('s3ql_metadata')
+        param = backend.lookup(meta_obj_name)
 
     # Check for unclean shutdown
     if param['seq_no'] < seq_no:
@@ -400,20 +407,14 @@ def get_metadata(backend, cachepath):
     elif param['max_inode'] > 2 ** 31:
         log.warning('Few free inodes remaining, running fsck is recommended')
 
-    if os.path.exists(cachepath + '-cache'):
-        for i in itertools.count():
-            bak_name = '%s-cache.bak%d' % (cachepath, i)
-            if not os.path.exists(bak_name):
-                break
-        log.warning('Found outdated cache directory (%s), renaming to .bak%d',
-                    cachepath + '-cache', i)
-        log.warning('You should delete this directory once you are sure that '
-                    'everything is in order.')
-        os.rename(cachepath + '-cache', bak_name)
-
     # Download metadata
     if not db:
         db = download_metadata(backend, cachepath + '.db')
+
+        # Drop cache
+        if os.path.exists(cachepath + '-cache'):
+            shutil.rmtree(cachepath + '-cache')
+
     save_params(cachepath, param)
 
     return (param, db)
@@ -430,7 +431,11 @@ def mark_metadata_dirty(backend, cachepath, param):
 def get_fuse_opts(options):
     '''Return fuse options for given command line options'''
 
-    fuse_opts = [ "nonempty", 'fsname=%s' % options.storage_url,
+    fsname=options.fs_name
+    if not fsname:
+        fsname = options.storage_url
+
+    fuse_opts = [ "nonempty", 'fsname=%s' % fsname,
                   'subtype=s3ql', 'big_writes', 'max_write=131072',
                   'no_remote_lock' ]
 
@@ -478,33 +483,17 @@ def parse_args(args):
                                      exitcode=35)
                 args.insert(pos, '--' + opt)
 
-    def compression_type(s):
-        hit = re.match(r'^([a-z0-9]+)(?:-([0-9]))?$', s)
-        if not hit:
-            raise argparse.ArgumentTypeError('%s is not a valid --compress value' % s)
-        alg = hit.group(1)
-        lvl = hit.group(2)
-        if alg not in ('none', 'zlib', 'bzip2', 'lzma'):
-            raise argparse.ArgumentTypeError('Invalid compression algorithm: %s' % alg)
-        if lvl is None:
-            lvl = 6
-        else:
-            lvl = int(lvl)
-        if alg == 'none':
-            alg =  None
-        return (alg, lvl)
-
     parser = ArgumentParser(
         description="Mount an S3QL file system.")
 
     parser.add_log('~/.s3ql/mount.log')
     parser.add_cachedir()
-    parser.add_authfile()
     parser.add_debug()
     parser.add_quiet()
     parser.add_backend_options()
     parser.add_version()
     parser.add_storage_url()
+    parser.add_compress()
 
     parser.add_argument("mountpoint", metavar='<mountpoint>', type=os.path.abspath,
                         help='Where to mount the file system')
@@ -516,6 +505,8 @@ def parse_args(args):
                       'this number you have to make sure that your process file descriptor '
                       'limit (as set with `ulimit -n`) is high enough (at least the number '
                       'of cache entries + 100).')
+    parser.add_argument("--keep-cache", action="store_true", default=False,
+                      help="Do not purge locally cached files on exit.")
     parser.add_argument("--allow-other", action="store_true", default=False, help=
                       'Normally, only the user who called `mount.s3ql` can access the mount '
                       'point. This user then also has full access to it, independent of '
@@ -527,15 +518,13 @@ def parse_args(args):
                            'user and the root user.')
     parser.add_argument("--fg", action="store_true", default=False,
                       help="Do not daemonize, stay in foreground")
-    parser.add_argument("--upstart", action="store_true", default=False,
-                      help="Stay in foreground and raise SIGSTOP once mountpoint "
-                           "is up.")
-    parser.add_argument("--compress", action="store", default='lzma-6',
-                        metavar='<algorithm-lvl>', type=compression_type,
-                        help="Compression algorithm and compression level to use when "
-                             "storing new data. *algorithm* may be any of `lzma`, `bzip2`, "
-                             "`zlib`, or none. *lvl* may be any integer from 0 (fastest) "
-                             "to 9 (slowest). Default: `%(default)s`")
+    parser.add_argument("--fs-name", default=None,
+                      help="Mount name passed to fuse, the name will be shown in the first "
+                           "column of the system mount command output. If not specified your "
+                           "storage url is used.")
+    parser.add_argument("--systemd", action="store_true", default=False,
+                      help="Run as systemd unit. Consider specifying --log none as well "
+                           "to make use of journald.")
     parser.add_argument("--metadata-upload-interval", action="store", type=int,
                       default=24 * 60 * 60, metavar='<seconds>',
                       help='Interval in seconds between complete metadata uploads. '
@@ -563,15 +552,12 @@ def parse_args(args):
     if options.allow_other and options.allow_root:
         parser.error("--allow-other and --allow-root are mutually exclusive.")
 
-    if not options.log and not options.fg:
+    if not options.log and not (options.fg or options.systemd):
         parser.error("Please activate logging to a file or syslog, or use the --fg option.")
 
     if options.profile:
         options.single = True
 
-    if options.upstart:
-        options.fg = True
-
     if options.metadata_upload_interval == 0:
         options.metadata_upload_interval = None
 
@@ -732,15 +718,11 @@ class CommitThread(Thread):
                 # ... number=500)/500 * 1e3
                 # 1.456586996000624
                 for el in list(self.block_cache.cache.values()):
-                    if self.stop_event.is_set():
+                    if self.stop_event.is_set() or stamp - el.last_write < 10:
                         break
-                    if stamp - el.last_access < 10:
-                        break
-                    if not el.dirty or el in self.block_cache.in_transit:
-                        continue
-
-                    self.block_cache.upload(el)
-                    did_sth = True
+                    if el.dirty and el not in self.block_cache.in_transit:
+                        self.block_cache.upload_if_dirty(el)
+                        did_sth = True
 
             if not did_sth:
                 self.stop_event.wait(5)
